/// @file   antsonruler.cpp
/// @brief  Simulation of ants walking on a ruler.
///
/// The ruler is subdivided into a grid. A number of ants is initially
/// put on the ruler, divided as evenly as possible over the grid.
/// Ants can move to a neighbouring grid point or stay put at every
/// time step.  How many of a square's ants move in each of the 3
/// different directions is random.  Ants can fall of the edges of the
/// ruler. The program prints the number of ants left on the ruler
/// after every time step, as well as the minimum number of ants on
/// any grid point and the maximum.
///
/// @author Ramses van Zon
/// @date   2021

#include <iostream>
#include <fstream>
#include <rarray>
#include "parameters.h"
#include "initialization.h"
#include "output.h"
#include "singlestep.h"
#include "ticktock.h"

/// @brief Main driver function of antsonruler
int main(int argc, char* argv[])
{
    // Parameters: take parameter file name from command line or use params.ini
    std::string inifilename("params.ini");
    if (argc >= 2) inifilename = argv[1];
    
    Parameters params = read_parameters(inifilename);
    print_parameters(params);

    // Record time
    TickTock tt;
    tt.tick();

    // Define arrays and other variables
    rvector<int> number_of_ants(params.length+2*nguards);     // distribution of ants on the ruler
    rvector<int> new_number_of_ants(params.length+2*nguards); // array used in time stepping
    int total_ants = params.totalants;     // varies in the simulation, but start with this many
    int nmin = params.totalants;           // will hold the minimum number of ants on any square
    int nmax = 0;                          // will hold the maximum number of ants on any square

    // open output file
    std::ofstream antsfile(params.antsfilename); // file stream

    tt.tock("Program start up took:");

    // Initialize system
    tt.tick();

    std::mt19937 engine(params.seed);
    initialize(number_of_ants, params.length, total_ants, nguards); // place the ants evenly on the ruler 
    minmaxtot(number_of_ants, nmin, nmax, total_ants);  // count ants and determine minimum and maximum number
    report(0, nmin, nmax, total_ants);
    write_array_binary(antsfile, number_of_ants, nguards);

    tt.tock("Initialization took:");

    // Simulation
    tt.tick();

    for (int timestep = 0; timestep < params.ntimesteps; timestep++) 
    {
        singlestep(number_of_ants, new_number_of_ants, engine);
        if ( (timestep+1) % params.outputinterval == 0) {
            minmaxtot(number_of_ants, nmin, nmax, total_ants);
            report(timestep+1, nmin, nmax, total_ants);
            write_array_binary(antsfile, number_of_ants, nguards);
        }
    }

    antsfile.close();

    tt.tock("Simulation took:");
}
