/// @file bin2nc.cpp
///
/// @brief Converts a very specific binary matrix format to netcdf format.
///
/// The expected binaray input format must consist of one of more
/// repeats of the following matrix structure:
///
///  nrow: 4 byte integer
///  ncols: 4 byte integer
///  followed by nrows*ncols data point of one specific numeric type.
///
/// Possible numeric types are byte, short int, int, and long int
/// (64-bits), their unsigned variants, and float and double.
/// The default type is int, but the type can be specified by the
/// third command line argument as "-b", "-s","-i", "-l", "-ub",
/// "-us", "-ui", "-ul", "-f" or "-d".
///
/// Subsequent structures with the same number of rows and columns
/// will be put in a three-dimensional variable in the netcdf file.
///
/// If the nrow is equal to 1, the matrix is flattened to a vector.
///
/// Ramses van Zon, Toronto, April 2021
///

#include <fstream>
#include <netcdf>
#include <cassert>
#include <string>

using std::string;
using std::to_string;
using std::ifstream;
using namespace netCDF;

// keep the detected structure of the binary file in a vector of the following

struct ArrayDescriptor {
    unsigned arr;      ///< array sequence number (starts at 1, used in the name of the variable)
    unsigned times;    ///< how many times repeated?
    int      rows;     ///< number of rows in this array (1 if one-dimensional data)
    int      columns;  ///< number of columns in this array
    NcVar    var;      ///< netcdf variable
};

// mapping elementary C++ types onto netcdf type (usage: "cpp_to_netcdf<TYPE>::type")

template<typename T> struct cpp_to_netcdf       { enum { type = 0 }; };
template<> struct cpp_to_netcdf<char>           { enum { type = NcType::ncType::nc_BYTE   }; };
template<> struct cpp_to_netcdf<short>          { enum { type = NcType::ncType::nc_SHORT  }; };
template<> struct cpp_to_netcdf<int>            { enum { type = NcType::ncType::nc_INT    }; };
template<> struct cpp_to_netcdf<ssize_t>        { enum { type = NcType::ncType::nc_INT64  }; };
template<> struct cpp_to_netcdf<unsigned char>  { enum { type = NcType::ncType::nc_UBYTE  }; };
template<> struct cpp_to_netcdf<unsigned short> { enum { type = NcType::ncType::nc_USHORT }; };
template<> struct cpp_to_netcdf<unsigned int>   { enum { type = NcType::ncType::nc_UINT   }; };
template<> struct cpp_to_netcdf<size_t>         { enum { type = NcType::ncType::nc_UINT64 }; };
template<> struct cpp_to_netcdf<float>          { enum { type = NcType::ncType::nc_FLOAT  }; };
template<> struct cpp_to_netcdf<double>         { enum { type = NcType::ncType::nc_DOUBLE }; };

// generic conversion function for any of the above types

template<typename T>
void convert(const string& binaryinname,
             const string& netcdfoutname)
{
    ifstream filein (binaryinname, std::ios::binary);  ///< binary input file stream
    NcFile   fileout(netcdfoutname, NcFile::replace);  ///< netcdf output file

    // read and interpret file structure
    std::vector<ArrayDescriptor> arrays;
    unsigned int arr = 0;
    int rows = 0;
    int columns = 0;
    while (filein.good()) {
        int nextrows;
        int nextcolumns;
        
        filein.read((char*)(&nextrows), sizeof(int));
        if (!filein.good()) break;
        assert(nextrows > 0);
            
        filein.read((char*)(&nextcolumns), sizeof(int));
        if (!filein.good()) break;        
        assert(nextcolumns > 0);
        
        filein.seekg(sizeof(T)*nextcolumns*nextrows, std::ios_base::cur);
        if (!filein.good()) break;
        
        if ( nextrows != rows || nextcolumns != columns ) {
            rows = nextrows;
            columns = nextcolumns;
            arrays.push_back({++arr,1,rows,columns});
        } else {
            arrays[arrays.size()-1].times ++;
        }
    }
    filein.close();

    // with that structure we can define the dimensions and variables
    for (auto& a: arrays) {
        std::vector<NcDim> dims;
        const string dimprefix("dim_");
        const string varprefix("arr_");
        // we will flatten dimensions of size 1, except for the columns.
        int dimnum = 0;
        if (a.times > 1) 
            dims.push_back(fileout.addDim(dimprefix+to_string(a.arr)+"_"+to_string(++dimnum),
                                          a.times));
        if (a.rows > 1)  
            dims.push_back(fileout.addDim(dimprefix+to_string(a.arr)+"_"+to_string(++dimnum),
                                          a.rows));
        dims.push_back(fileout.addDim(dimprefix+to_string(a.arr)+"_"+to_string(++dimnum),
                                      a.columns));
        a.var = fileout.addVar(varprefix + to_string(a.arr), cpp_to_netcdf<T>::type, dims);
    }
    
    // write data
    filein.open(binaryinname,std::ios::binary);
    for (auto& a: arrays) {
        int* buffer = new int[a.times*a.rows*a.columns];
        for (unsigned int t = 0; t < a.times; t++) {
            int checkrows, checkcolumns;
            filein.read((char*)(&checkrows),    sizeof(int));
            filein.read((char*)(&checkcolumns), sizeof(int));
            // consistency check
            assert(checkrows==a.rows);
            assert(checkcolumns==a.columns);
            filein.read((char*)(buffer+t*a.rows*a.columns), sizeof(T)*a.rows*a.columns);
        }      
        a.var.putVar(buffer);
        delete[] buffer;
    }
    filein.close();
}

    
int main(int argc, char** argv)
{
    // Interpret command line arguments
    //
    // argument 1: input binary file
    //
    // argument 2: output netcdf file (optional)
    //
    // argument 3: type of data: -b for byte/char, -s for short int,
    //             -i for int, -l for long long int, -f for float,
    //             -d for double, -ub  for unsigned int byte,
    //             -us for unsigned short, -ui for unsigned int,
    //             -ul for unsigned long long int.

    if ( (argc<=1) || (argc>=5)) {
        std::cerr << "Usage:\n\n"
                  << "   " << argv[0] << " BINARYFILE [OUTPUTNETCDFFILE] [-TYPE]\n\n";
        return 1;
    }
    
    string namein(argv[1]);
    string nameout((argc>2)?argv[2]:(string(argv[1]) + ".nc"));
    string type((argc>3)?argv[3]:"-i");

    // call a different instantiation for each type
    switch (type[1]) {
      case 'b': convert<char>(namein, nameout); break;
      case 'i': convert<int>(namein, nameout); break;
      case 's': convert<short>(namein, nameout); break;
      case 'l': convert<ssize_t>(namein, nameout); break;
      case 'f': convert<float>(namein, nameout); break;
      case 'd': convert<double>(namein, nameout); break;
      case 'u' : { switch (type[2]) {
                     case 'b': convert<unsigned char>(namein, nameout); break;
                     case 'i': convert<unsigned int>(namein, nameout); break;
                     case 's': convert<unsigned short>(namein, nameout); break;
                     case 'l': convert<size_t>(namein, nameout); break;
                   }
                   break;
                 } 
    }
}
