/// @file   randompartitioning.cpp
/// @brief  Randomly separate numbers into sets
/// @author Ramses van Zon
/// @date   2021

#include <random>
#include <algorithm>

#include "randompartitioning.h"

/// @brief Function to randomly divide a number 'total' into a set of
///        'numdivision' numbers that add up to 'total'. Uses the C++
///        random library to generate the randomness.
///
/// @param  engine    of type mt19937, to produce random numbers
/// @param  total     number that is to be subdivided (input)
/// @param  nparts    number of partitions into which to divide 'total (input)
/// @param  nperpart  resulting number in each partition (output)
///
/// @return void
///
void rand_partition(std::mt19937& engine, int total, int nparts, int* nperpart)
{
    // Draw the points at which to 'split' the number
    std::uniform_int_distribution<int> uniformint(0, total);
    nperpart[0] = 0;
    for (int k = 1; k < nparts; k++)
        nperpart[k] = uniformint(engine);
    // By sorting, the size of each part is the just difference of
    // subsequent split points
    std::sort(&nperpart[0], &nperpart[nparts]);
    for (int k = 0; k < nparts-1; k++)
        nperpart[k] = nperpart[k+1] - nperpart[k];
    nperpart[nparts-1] = total - nperpart[nparts-1];
}

